#!/usr/bin/env node
// Nginx management process.
//
// All this does is figure out where to find the proxy's nginx config, then
// attempt to start it.  restarts are the responsibility of systemd.

const { spawn } = require('child_process');
const path = require('path');
const fs = require('fs');

let child = null;

async function readJSONFile(filepath) {
  let filecontents = "";
  let obj = {};
  try {
    filecontents = await fs.promises.readFile(filepath, { encoding: 'utf8' });
    obj = JSON.parse(filecontents);
  } catch (e) {
    return undefined;
  }
  return obj;
}

async function getVarPrefix() {
  const crconf = process.env.CORONERD_CONF || '/etc/coronerd/coronerd.conf';
  const crjson = await readJSONFile(crconf);
  if (crjson && crjson.varprefix) {
    return crjson.varprefix;
  }
  return '/var/run/coronerd';
}

function shutdown() {
  /* forward shutdown signal from parent to child. */
  if (child) {
    child.kill('SIGTERM');
  }
  process.exit(0);
}

function failure() {
  /* force child to exit and exit with error. */
  if (child) {
    child.kill('SIGTERM');
  }
  process.exit(1);
}

function sigchld() {
  /* do nothing; should be handled by on("exit"). */
}

function reload() {
  /* forward reload signal from parent to child. */
  if (child) {
    child.kill('SIGHUP');
  }
}

async function do_mkdir(dirpath) {
  const parent = path.dirname(dirpath);
  if (parent !== dirpath) {
    await do_mkdir(parent);
  }
  try {
    await fs.promises.mkdir(dirpath);
  } catch (e) {
    if (e.code !== 'EEXIST')
      throw e;
  }
}

async function mkdirp(dirpath) {
  const normdirpath = path.normalize(dirpath);
  return do_mkdir(normdirpath);
}

async function sleep(delay_ms) {
  return new Promise((resolve) => { setTimeout(resolve, delay_ms); });
}

async function wait_for_file(filepath) {
  for (;;) {
    try {
      await fs.promises.stat(filepath);
      break;
    } catch (e) {
      await sleep(500);
    }
  }
}

async function main(argv) {
  const shutdownsigs = "SIGINT SIGTERM".split(" ");
  const failuresigs = "SIGPIPE SIGUSR1 SIGUSR2 SIGALRM".split(" ");
  const varprefix = await getVarPrefix();
  const wdir = process.env.BT_NGINX_WDIR || `${varprefix}/backtrace-proxy`;
  const nconf = process.env.BT_NGINX_CONF || `${wdir}/nginx.conf`;
  const nginx = process.env.BT_NGINX_PATH || '/usr/sbin/nginx';
  const nginx_args = [
    '-p', wdir,
    '-c', nconf,
    '-g', 'error_log error.log;'
  ];

  /* Fulfill prerequisites before starting nginx. */
  await mkdirp(wdir);
  process.chdir(wdir);
  console.log(`Waiting for ${nconf} to appear...`);
  await wait_for_file(nconf);

  const nginx_argv = nginx_args.map((i) => {
    return (i.indexOf(" ") == -1) ? i : `"${i}"`;
  });
  console.log(`Starting: nginx ${nginx_argv.join(" ")}`);
  child = spawn(nginx, nginx_args, {
    stdio: 'inherit',
  });
  shutdownsigs.forEach((sig) => {
    process.on(sig, shutdown);
  });
  failuresigs.forEach((sig) => {
    process.on(sig, failure);
  });
  process.on('SIGCHLD', sigchld);
  process.on('SIGHUP', reload);
  child.on('exit', (code, signal) => {
    console.log(`Nginx child exited with code ${code} signal ${signal}`);
    if (code === null)
      process.exit(1);
    process.exit(code);
  });
}

main(process.argv.slice(2));
// vim:ts=2:sw=2:et
