#!/usr/bin/env node
// Basic proxy server implementation.
//
// This works by monitoring a descriptor file specified as the first argument,
// which is written out periodically whenever coronerd has updates.  The
// second argument specifies a directory this program can write to.
//
// The descriptor contents look like this:
// {
//   "bind": [
//     {
//       "hostname": "0.0.0.0",
//       "service": "4567"
//     }
//   ],
//   "ssl": {
//     "certificate_chain_file": "/etc/coronerd/ssl/chain.crt",
// OR  "certificate": "/etc/coronerd/ssl/foo.crt",
//     "key": "/etc/coronerd/ssl/sp.backtrace.io.key"
//   },
//   "proxy": {
//     "default": "http://localhost:8080",
//     "/api/simple": "http://localhost:8081",
//     "/api/saml": "http://localhost:8082",
//     [...]
//   }
// }
//
// The expectation is that every proxy target accepts that only requests for
// their specific subset of the full namespace will be forwarded to them.
// The "default" proxy target gets everything else.
//
// Additionally, to remove the need for any npm modules, this script uses
// only built-in node libraries.
//
// To allow for both a dedicated proxy server, and for integration with a
// system nginx service, separate files are written out for both a top-level
// nginx configuration and a service-specific configuration.  Whenever these
// are updated, the server pid file is used to notify the nginx service.
//

const fs        = require('fs');
const net       = require('net');
const path      = require('path');
const util      = require('util');
const cp        = require('child_process');

var descriptor = null;
var last_update = undefined;
var cur_update = undefined;
var watcher = null;
var proxy_prefix = null;
var nginx_pidfile = null;
var nginx_cfgfile = null;
var nginx_vhfile = null;
var nginx_temp_paths = [];
var logstream = null;
var inited = false;
var startup_ts = null;

function log(level, str) {
  if (level === 'debug' && !process.env.BT_PROXY_DEBUG)
    return;
  if (logstream)
    logstream.write(`${Date()} cr_proxy.js[${level}] ${str}\n`);
}

function nginx_print(str, indent) {
  return " ".repeat(indent) + str + "\n";
}

function obj_to_s(obj) {
  return JSON.stringify(obj, null, 4);
}

function notify_nginx() {
  var pid = parseInt(fs.readFileSync(nginx_pidfile));
  log("debug", `Sending SIGHUP to ${pid} ...`);
  process.kill(pid, 'SIGHUP');
}

async function readJSONFile(filepath) {
  let filecontents = "";
  let obj = {};
  try {
    filecontents = await fs.promises.readFile(filepath, { encoding: 'utf8' });
    obj = JSON.parse(filecontents);
  } catch (e) {
    return undefined;
  }
  return obj;
}

/*
 * Convert an object into an nginx config file.  This uses a key-value pair
 * object model, with keys being config items, and values being their values.
 * A config item may have multiple values, in which case the value for the
 * key should be an array of all the values to be printed.
 *
 * Misc notes:
 * - nginx sections are always objects in this representation.
 * - POD numbers may be values, but are treated as strings in translation.
 * - e.g. "location x" is not an array, since the location parameters are
 *   part of the key.
 */
function obj2nginxcfg(obj, indent) {
  var cfg = "";

  if (!indent)
    indent = 0;

  for (key in obj) {
    const val = obj[key];

    if (Array.isArray(val)) {
      val.forEach((elem) => {
        if (typeof(elem) !== 'string' && typeof(elem) !== 'number')
          throw new Error("nginx array values must be POD");
        cfg += nginx_print(key + " " + elem + ";", indent);
      });
      continue;
    }

    if (typeof(val) === 'string' || typeof(val) === 'number') {
      cfg += nginx_print(key + " " + val + ";", indent);
      continue;
    }

    cfg += nginx_print(key + " {", indent);
    cfg += obj2nginxcfg(val, indent + 2);
    cfg += nginx_print("}", indent);
  }
  return cfg;
}

/*
 * Helper that returns a promise which resolves if the port given is
 * available for use.  Any error is treated as a reject, while a listening
 * event is treated as a resolve, once the port closes again.
 */
const port_available = (port) => new Promise((resolve, reject) => {
  const srv = net.createServer()
  .once('error', err => reject(err))
  .once('listening', () => {
    const srv_port = srv.address().port;
    srv.once('close', () => resolve(srv_port)).close();
  })
  .listen(port)
});

/*
 * Helper that returns a promise which resolves with a port number once an
 * available port number is found.  A preferred port may be used to start.
 */
const get_port = (port) => new Promise((resolve, reject) => {
  log("debug", `trying port ${port} ...`);
  return port_available(port).then((port) => {
    log("debug", `resolved with port ${port} ...`);
    resolve(port);
  }).catch((err) => {
    /* Try again and let the kernel give us an usable port. */
    log("debug", `got err ${err}`);
    port_available(0).then((port) => {
      log("debug", `backup resolved with port ${port} ...`);
      resolve(port);
    }).catch((err) => {
      log("debug", `backup err: ${err} ...`);
      reject(err);
    });
  });
});

function stat_equal(st0, st1) {
  return st0.inode === st1.inode &&
    st0.size === st1.size &&
    st0.mtimeMs === st1.mtimeMs &&
    st0.ctimeMs === st1.ctimeMs;
}

function nginx_config_opts() {
  const nginx_path = process.env.NGINX_PATH || "/usr/sbin/nginx";
  let params = [];
  try {
    const result = cp.execSync(`${nginx_path} -V 2>&1`).toString();
    const cfg = result.split("\n").find((s) => {
      return s.startsWith("configure arguments:");
    });
    params = cfg.split("--");
    params.shift();

    return params.reduce((accum, val) => {
      const kv = val.split("=");
      if (kv.length === 2 && kv[0].endsWith("-temp-path")) {
        accum.push(kv[0].replace("http-", "").replace(/-/gi, "_"));
      }
      return accum;
    }, []);
  } catch (e) {
    log("error", `Unable to query nginx vars: ${e}`);
    log("error", e.stdout);
  }
  return [];
}

async function stat(fname) {
  try {
    const st = await fs.promises.lstat(fname);
    return st;
  } catch (e) {
  }
  return null;
}

/*
 * Get the _all.json descriptor configuration.  This returns:
 * - null: The descriptor is unchanged since the last check.
 * - object: The descriptor exists and requires updating.
 */
async function get_descriptor_config() {
  /*
   * Determine if an update has actually occurred.  There may be multiple
   * pending calls to this function at once, so exclude duplicates.  If this
   * throws an exception, the timer will try again in a second, in case the
   * file is missing or has a parse error.
   *
   * Note that coronerd will change startup_ts whenever it restarts, so
   * this is treated as a change event.
   */
  const desc = await readJSONFile(descriptor);
  const old_ts = startup_ts;

  cur_update = desc.generation;
  if (startup_ts !== null && startup_ts === desc.startup_ts) {
    if (last_update !== undefined && cur_update === last_update) {
      log("debug", `descriptor appears unchanged (${desc.startup_ts}+${cur_update}) since last check (${startup_ts}+${last_update})`);
      return null;
    }
  } else {
    log("debug", `startup_ts changed (was ${startup_ts} now ${desc.startup_ts})`);
    startup_ts = desc.startup_ts;
  }
  log("debug",
    `descriptor is now ${old_ts}+${cur_update}, was ${startup_ts}+${last_update}`);
  return desc;
}

async function update_config() {
  var proxy = null;
  var listen_add = "";
  /* Nginx must expect to be run with `-p`, so all paths are relative. */
  var nginx_top = {
    daemon: 'off',
    pid: nginx_pidfile,
    events: {
      worker_connections: 2000,
    },
    http: {
      client_max_body_size: 0,
      proxy_read_timeout: "10m",
      include: `${proxy_prefix}/_proxy.nginx.conf.http.d/*`,
      access_log: 'access.log combined',
      server: {
        listen: [],
        error_page: "500 502 503 504 /status/error.html",
        "location /status": {
          ssi: "on",
          auth_basic: "off",
          root: `${proxy_prefix}/www`,
        },
      },
    },
    include: `${proxy_prefix}/_proxy.nginx.conf.d/*`,
  };
  var new_cfg = null;
  var err = null;

  /* Shortcut var as most stuff goes in this section... */
  var server = nginx_top.http.server;

  /* attempt to override all paths that might be needed. */
  nginx_temp_paths.forEach((tmppath) => {
    nginx_top.http[tmppath] = tmppath.replace("_path", "");
  });

  log("debug", "processing configuration...");
  cur_update = undefined;

  try {
    new_cfg = await get_descriptor_config();
    if (new_cfg === null) {
      return;
    }
  } catch (e) {
    log("error", `failed to process descriptor: ${e}`);
    new_cfg = undefined;
  }

  if (!new_cfg || !new_cfg.proxy || !cur_update) {
    /*
     * If the descriptor is unavailable or does not parse correctly, determine
     * whether it is necessary to create a stub nginx configuration, so
     * that backtrace-nginx can come up.
     */
    try {
      await stat(nginx_cfgfile);
      /* If the call succeeds, the nginx.conf file exists, don't change it. */
      log("debug", "descriptor is missing or incomplete");
      return;
    } catch (st_e) {
      log("info", "creating stub configuration");
      new_cfg = {};
    }
  }

  /* Process the configuration to see if it's actionable. */
  try {
    proxy = new_cfg.proxy;
    ssl = new_cfg.ssl;

    if (new_cfg.ssl) {
      const chain_file = new_cfg.ssl.certificate_chain_file;
      const cert_file = new_cfg.ssl.certificate_file;
      const ssl_key = new_cfg.ssl.key;
      if (!ssl_key)
        throw new Error("Must specify SSL key file");
      if (chain_file) {
        server.ssl_certificate = chain_file;
      } else {
        if (!cert_file)
          throw new Error("Must specify SSL certificate or certificate chain");
        server.ssl_certificate = cert_file;
      }
      server.ssl_certificate_key = ssl_key;
      listen_add = " ssl";
      server.ssl_protocols = "TLSv1 TLSv1.1 TLSv1.2";
      server.ssl_ciphers = "HIGH:!aNULL:!MD5";
    }

    if (new_cfg.bind) {
      new_cfg.bind.forEach((bindobj) => {
        server.listen.push(bindobj.service + listen_add);
      });
    } else {
      const actual_port = await get_port(443);
      server.listen.push(actual_port + listen_add);
    }

    if (proxy) {
      for (key in proxy) {
        server["location " + key] = {
          proxy_pass: proxy[key],
          proxy_set_header: ["X-Proxied-For $remote_addr:"],
          proxy_buffering: "off",
        };
        if (key === '/') {
          /* coronerd needs this to resolve the universe */
          server["location /"].proxy_set_header.push("Host $host");
        }
      }
    } else {
      log("debug", "empty descriptor, stubbing");
      server["location /"] = {
        proxy_pass: "http://localhost",
        proxy_buffering: "off",
      };
    }

    /* Write out updated configuration and mv into place. */
    const nginx_vhtmp = nginx_vhfile + ".tmp";
    const nginx_cftmp = nginx_cfgfile + ".tmp";
    await fs.promises.writeFile(nginx_vhtmp, obj2nginxcfg(nginx_top.http));
    await fs.promises.writeFile(nginx_cftmp, obj2nginxcfg(nginx_top));
    await fs.promises.rename(nginx_vhtmp, nginx_vhfile);
    await fs.promises.rename(nginx_cftmp, nginx_cfgfile);

    notify_nginx();
  } catch (e) {
    err = e;
  }

  log("info", `${nginx_cfgfile} (ts ${startup_ts} gen ${cur_update}) updated to: ${obj_to_s(nginx_top)}`);
  last_update = cur_update;
  return;
}

function logpath() {
  if (process.env.BT_PROXY_LOGPATH)
    return process.env.BT_PROXY_LOGPATH;
  if (process.env.BT_PROXY_LOGNAME)
    return `${proxy_prefix}/${process.env.BT_PROXY_LOGNAME}`;
  return null;
}

async function getVarPrefix() {
  const crconf = process.env.CORONERD_CONF || '/etc/coronerd/coronerd.conf';
  const crjson = await readJSONFile(crconf);
  if (crjson && crjson.varprefix) {
    return crjson.varprefix;
  }
  return '/var/run/coronerd';
}

async function do_mkdir(dirpath) {
  const parent = path.dirname(dirpath);
  if (parent !== dirpath) {
    await do_mkdir(parent);
  }
  try {
    await fs.promises.mkdir(dirpath);
  } catch (e) {
    if (e.code !== 'EEXIST')
      throw e;
  }
}

async function mkdirp(dirpath) {
  const normdirpath = path.normalize(dirpath);
  return do_mkdir(normdirpath);
}

async function delay(timeout_ms) {
  return new Promise(resolve => setTimeout(resolve, timeout_ms));
}

async function cp_dir(src_dir, dest_dir) {
  if (fs.existsSync(src_dir) === false)
    return;
  const dirents = await fs.promises.readdir(src_dir);
  return Promise.all(dirents.map((de) => {
    return fs.promises.copyFile(`${src_dir}/${de}`, `${dest_dir}/${de}`);
  }));
}

async function main(argv) {
  /*
   * Process command line arguments.
   */
  if (argv[0] === '-h' || argv[0] === '--help') {
    console.log(`Usage: ${process.argv[1]} <descriptor> <var prefix> <nginx.pid>`);
    process.exit(1);
  }

  /* Initialize context by updating from the descriptor. */
  descriptor = argv[0] || process.env.BT_PROXY_DESCRIPTOR;
  proxy_prefix = argv[1] || process.env.BT_PROXY_PREFIX;
  nginx_pidfile = argv[2] || process.env.BT_PROXY_NGINX_PIDFILE;
  const varprefix = await getVarPrefix();

  if (!descriptor)
    descriptor = `${varprefix}/svclayer/_all.json`;
  if (!proxy_prefix)
    proxy_prefix = `${varprefix}/backtrace-proxy`;
  if (!nginx_pidfile)
    nginx_pidfile = `${proxy_prefix}/_proxy.nginx.pid`;
  nginx_cfgfile = `${proxy_prefix}/nginx.conf`;
  nginx_vhfile = `${proxy_prefix}/nginx.vhost.conf`;
  nginx_temp_paths = nginx_config_opts();

  await mkdirp(`${proxy_prefix}/www/status`);
  await cp_dir("/opt/backtrace/coronerd/status", `${proxy_prefix}/www/status`);

  process.chdir(proxy_prefix);
  const logfilepath = logpath();
  if (logfilepath) {
    logstream = fs.openSync(logfilepath, 'a');
    logstream.write("\n----\n");
  } else {
    logstream = process.stdout;
  }

  log('info', `Started, will monitor ${descriptor}, pidfile ${nginx_pidfile}`);
  log('debug', `nginx temp paths: ${JSON.stringify(nginx_temp_paths, null, 4)}`);

  for (;;) {
    await update_config();
    await delay(1000);
  }
}

main(process.argv.slice(2));

// vim:ts=2:sw=2:et
