-- Example usage:
-- ptrace --sandbox-whitelist=external_modules --load= --global= <coronerpid>

local bit32 = require "bit32"

-- These global variables are used to accumulate multiple fields in the final
-- JSON annotation.
json_fields = {}
smr_fields = {}
event_count = 0
event_count_type = {}

local function eventlist_to_string(t, event_ptr_bt, status_str)
	local event_ptr_pt = pmodule.variable_from_bt(event_ptr_bt)

	while event_ptr_pt:value() ~= 0 do
		local event_bt = pmodule.deref(event_ptr_bt)
		event_count_type[status_str] = event_count_type[status_str] + 1
		event_count = event_count + 1
		event_ptr_bt = event_bt.floatpin_linkage.cle_next
		event_ptr_pt = pmodule.variable_from_bt(event_ptr_bt)
	end
end

local function smr_sections_to_string(t, section_bt)
	local section_pt = pmodule.variable_from_bt(section_bt)
	local n = 0

	while section_pt:value() ~= 0 do
		local fn_bt = pmodule.variable_from_bt(pmodule.deref(section_bt.fn))
		local str = string.format('"%d": "%s:%d"', n, fn_bt:symbol(),
		    pmodule.variable_from_bt(section_bt.line):value())

		table.insert(t, str)
		section_bt = section_bt.entry.csle_next
		section_pt = pmodule.variable_from_bt(section_bt)
		n = n + 1
	end
end

-- Add per-thread SMR metadata to the table thread_t.
-- Add per-thread active entries to the global table smr_fields.
local function smr_to_string(threadid, thread_t, smr_bt)
	local sections = {}
	local sections_string = nil
	local fn_bt = pmodule.variable_from_bt(pmodule.deref(smr_bt.fn_highwater))

	smr_sections_to_string(sections, smr_bt.active.cslh_first)
	if next(sections) ~= nil then
		sections_string = table.concat(sections,",")
	end

	local count = pmodule.variable_from_bt(smr_bt.highwater):value()

	if (count > 0) then
		local str = string.format(
		    '"SMR": { "Highwater" : { "Count": %d, "Location": "%s:%d" }',
		    count,
		    fn_bt:symbol(),
		    pmodule.variable_from_bt(smr_bt.line_highwater):value())

		if (sections_string ~= nil) then
			str = str ..
				string.format(', "Sections" : { %s }',
				    sections_string)
		end
		str = str .. '}'
		table.insert(thread_t, str)
	end
end

-- current_pt is a "ptrace variable" representing the current ptrace state.
-- current_bt is a "bt variable" representing a PU object.
-- both vars are "reference variable" because of the match criterion
local function current_var_cb(current_pt, current_bt)
	local pu_strings = {}
	local thread_events_str
	local threadid = current_pt:thread():tid()

	event_count_type["pinned"] = 0
	eventlist_to_string(pu_strings,
	    current_bt.notify.io.pinned_events.clh_first, "pinned")
	event_count_type["floated"] = 0
	eventlist_to_string(pu_strings,
	    current_bt.notify.io.floated_events.clh_first, "floated")

	smr_to_string(threadid, pu_strings, current_bt.smr)

	local name = pmodule.variable_from_bt(current_bt.name):value()
	name = pmodule.address_read_string(name, 256)

	thread_events_str = string.format('"%s (%s)" : { %s }',
	    name, threadid, table.concat(pu_strings, ", "))
	table.insert(json_fields, thread_events_str)
end

function postattach_cb(x)
	-- pmodule.log(pmodule.log_level.warning, "json_str: %s", json_str)
	pmodule.backtrace():annotate(pmodule.annotation.json,
	    '{ "json" : { "Threads": { %s } } }',
	    table.concat(json_fields, ", "))

	pmodule.backtrace():add_kv_string("events", event_count);
	pmodule.backtrace():add_kv_string("events.pinned",
	    event_count_type["pinned"]);
	pmodule.backtrace():add_kv_string("events.floated",
	    event_count_type["floated"]);
end

function pm_load()
	-- Select only variables named "current" from the filename "pu.c"
	local m = pmodule.match()
	m:add_variable_name("current", pmodule.match_type.substr)
	m:add_variable_ptrace_type(pmodule.variable_type.reference)
	m:add_file("pu.c", pmodule.match_type.substr)

	pmodule.register(pmodule.event.variable, current_var_cb, m)
	pmodule.register(pmodule.event.postattach, postattach_cb)
end

pmodule.define{id = "coronerd_lua",
               load = pm_load}

